{
===============
 JSTable Notes
===============


DESCRIPTION
-----------
JSTable is a TTable descendant that is able to use dBase expression indexes
in Locate/Lookup (the standard TTable implementation will never select
an expression index to do the search).

There are many more areas in TTable that could be improved this way.
In fact, the standard Delphi implementation of TTable is written without
taking expression indexes into account at all.

Things that should still be enhanced:
- FindKey/FindNearest
- GotoKey/GotoNearest
- SetRange (SetRangeStart/SetRangeEnd/ApplyRange works OK)
- MasterSource/MasterFields (master/detail)

Since none of these methods/properties are declared virtual, I didn't
implement them.  That's because if the incorrect cast (TTable instead
of TJSTable) were used on an object, the wrong methods/properties would
be addressed.

This component can be used with Delphi 3 and 4.


COPYRIGHT/WARANTY
-----------------
This component is freeware.  You may use it in compiled form in your programs.
You may alter it and distribute it further in source code as long as this is
done without commercial purposes and on the condition that this comment is
included integrally in it.

The author doesn't warrant anything and doesn't accept liability for any
possible damage or data loss this component might cause.  You use this
component entirely at your own risk.

If you have questions or suggestions regarding this component, you can
contact me via email:
    Jan Sprengers
    jan@adm2000.be

}

unit JSTbl;

interface

uses
  Windows, Messages, SysUtils, Classes, DB, DBTables, DBCommon, DBExcept, BDE;

type
  PFieldDescList = ^TFieldDescList;
  TFieldDescList = array[0..1023] of FLDDesc;

  TDBaseExprInfo = record
    Expression: String;
    KeyLength: Integer;
  end;

  TJSTable = class(TTable)
  private
    { Private declarations }
    FLookupHandle: HDBICur;
    FLookupIndexExpr: String;
    FInLookup: Boolean;
    FFieldDescs: PFieldDescList;
    FFieldDescCount: Integer;
  protected
    { Protected declarations }
    procedure DestroyHandle; override;
    function JSLocateRecord(const KeyFields: String;
      const KeyValues: Variant; Options: TLocateOptions;
      SyncCursor: Boolean): Boolean;
    function JSGetLookupCursor(const SearchExpr: String): HDBICur;
    procedure JSDestroyLookupCursor;
    procedure InitFieldDefs; override;
    function GetIsIndexField(Field: TField): Boolean; override;
    procedure DestroyLookupCursor; override;
    procedure GetDBaseExprInfo(const Fields: String; Options: TLocateOptions;
      var ExprInfo: TDBaseExprInfo);
    procedure GetDBaseFieldExprInfo(const FieldName: String;
      var ExprInfo: TDBaseExprInfo);
    function GetFieldDesc(FieldName: String): PFldDesc;
  public
    { Public declarations }
    procedure AddDBaseIndex(const Name, Fields: String; Options: TIndexOptions;
      Maintained: Boolean);
    function Locate(const KeyFields: String; const KeyValues: Variant;
      Options: TLocateOptions): Boolean; override;
    function Lookup(const KeyFields: String; const KeyValues: Variant;
      const ResultFields: string): Variant; override;
  end;

procedure Register;

implementation

{ JSTable }

// Allows creation of indexes for dBase tables.
// You can use the syntax for normal indexes (fields separated by ';')
// and thus ensure that your index will be compatible for Locate/Lookup
procedure TJSTable.AddDBaseIndex(const Name, Fields: string;
  Options: TIndexOptions; Maintained: Boolean);
var
  FieldExpr: String;
  ExprInfo: TDbaseExprInfo;
  ExprOptions: TLocateOptions;
  IndexDesc: IDXDesc;
  IdxName: DbiTblName;
begin
  FieldDefs.Update;
  if (Pos(';', Fields) > 0) or (ixCaseInsensitive in Options) then
  begin
    // Convert field-based index request to expression index
    if ixCaseInsensitive in Options then
      ExprOptions := [loCaseInsensitive] else
      ExprOptions := [];
    GetDBaseExprInfo(Fields, ExprOptions, ExprInfo);
    FieldExpr := ExprInfo.Expression;
    Options := Options + [ixExpression] - [ixCaseInsensitive];
  end else
    FieldExpr := Fields;
  if not Maintained then
  begin
    // Allow non-maintained (.NDX-style) indexes
    EncodeIndexDesc(IndexDesc, '', FieldExpr, Options);
    StrLCopy(@IdxName, PChar(Name), SizeOf(IdxName));
    IndexDesc.bMaintained := False;
    IndexDesc.szName := IdxName;
  end else
    EncodeIndexDesc(IndexDesc, Name, FieldExpr, Options);
  if Active then
  begin
    CheckBrowseMode;
    CursorPosChanged;
    Check(DbiAddIndex(DBHandle, Handle, nil, nil, IndexDesc, nil));
  end else
  begin
    SetDBFlag(dbfTable, True);
    try
      Check(DbiAddIndex(DBHandle, nil, PChar(TableName), szDBase,
        IndexDesc, nil));
    finally
      SetDBFlag(dbfTable, False);
    end;
  end;
  IndexDefs.Updated := False;
end;

procedure TJSTable.DestroyHandle;
begin
  // Bug in Delphi: Delphi releases the record lock on the current record.
  // If the record is locked on another table object in the same session,
  // that lock is released.  We avoid this by placing the cursor at EOF.
  if State <> dsEdit then
    DbiSetToEnd(Handle);
  // Cleanup
  if FFieldDescs <> nil then
  begin
    FreeMem(FFieldDescs, FFieldDescCount * SizeOf(FLDDesc));
    FFieldDescs := nil;
  end;
  inherited;
end;

// Actually it is LocateRecord that we want to modify.
// Unfortunately, LocateRecord is not virtual; Lookup is.
// Code copied from VCL source and LocateRecord replaced with JSLocateRecord.
function TJSTable.Lookup(const KeyFields: string; const KeyValues: Variant;
  const ResultFields: string): Variant;
begin
  Result := Null;
  if JSLocateRecord(KeyFields, KeyValues, [], False) then
  begin
    SetTempState(dsCalcFields);
    try
      CalculateFields(TempBuffer);
      Result := FieldValues[ResultFields];
    finally
      RestoreState(dsBrowse);
    end;
  end;
end;

// Actually it is LocateRecord that we want to modify.
// Unfortunately, LocateRecord is not virtual; Locate is.
// Code copied from VCL source and LocateRecord replaced with JSLocateRecord.
function TJSTable.Locate(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions): Boolean;
begin
  DoBeforeScroll;
  Result := JSLocateRecord(KeyFields, KeyValues, Options, True);
  if Result then
  begin
    Resync([rmExact, rmCenter]);
    DoAfterScroll;
  end;
end;

// Special lookup for dBase expression indexes
function TJSTable.JSLocateRecord(const KeyFields: String;
  const KeyValues: Variant; Options: TLocateOptions;
  SyncCursor: Boolean): Boolean;
var
  ExprInfo: TDBaseExprInfo;
  ItrFld, FieldCount, KeyLength: Integer;
  Fields: TList;
  LookupCursor: HDBICur;
  Filter: HDBIFilter;
  Status: DBIResult;
  CursorProps: CurProps;
  KeyBuffer: PKeyBuffer;
  IndexBuffer, RecBuffer: PChar;
  UseKey: Boolean;
begin
  if (TableType = ttDBase) or ((TableType = ttDefault) and
     (CompareText(ExtractFileExt(TableName), '.DBF') = 0)) then
  begin
    // Special handling for dBase expression indexes
    CheckBrowseMode;
    CursorPosChanged;
    Fields := TList.Create;
    try
      GetFieldList(Fields, KeyFields);
      if (loPartialKey in Options) and
         (TField(Fields.Last).DataType <> ftString) then
        Options := Options - [loPartialKey];
      if not CachedUpdates then
      begin
        GetDBaseExprInfo(KeyFields, Options, ExprInfo);
        LookupCursor := JSGetLookupCursor(ExprInfo.Expression);
      end else
        LookupCursor := nil;
      if LookupCursor <> nil then
      begin
        DbiGetCursorProps(LookupCursor, CursorProps);
        FInLookup := True;
        SetTempState(dsFilter);
        SetKeyBuffer(kiLookup, True);
        IndexBuffer := AllocMem(CursorProps.iKeySize);
        try
          FieldCount := Fields.Count;
          if FieldCount = 1 then
          begin
            TField(Fields.First).Value := KeyValues;
          end else
          begin
            for ItrFld := 0 to FieldCount - 1 do
              TField(Fields[ItrFld]).Value := KeyValues[ItrFld];
          end;
          KeyLength := ExprInfo.KeyLength;
          if loPartialKey in Options then
            KeyLength := KeyLength + Length(TField(Fields.Last).AsString);
          KeyBuffer := GetKeyBuffer(kiLookup);
          RecBuffer := PChar(KeyBuffer) + SizeOf(TKeyBuffer);
          UseKey := DbiExtractKey(LookupCursor, RecBuffer, IndexBuffer) = 0;
          if UseKey then
            RecBuffer := IndexBuffer;
          Status := DbiGetRecordForKey(LookupCursor, UseKey, 0, KeyLength,
            RecBuffer, TempBuffer);
        finally
          FreeMem(IndexBuffer, CursorProps.iKeySize);
          RestoreState(dsBrowse);
          FInLookup := False;
        end;
        if (Status = DBIERR_NONE) and SyncCursor and
           (LookupCursor <> Handle) then
          Status := DbiSetToCursor(Handle, LookupCursor);
      end else
      begin
        Check(DbiSetToBegin(Handle));
        Filter := CreateLookupFilter(Fields, KeyValues, Options, 2);
        DbiActivateFilter(Handle, Filter);
        Status := DbiGetNextRecord(Handle, dbiNoLock, TempBuffer, nil);
        DbiDropFilter(Handle, Filter);
      end;
    finally
      Fields.Free;
    end;
    Result := Status = DBIERR_NONE;
  end else
    Result := LocateRecord(KeyFields, KeyValues, Options, SyncCursor);
end;

// Gets a temporary cursor for an index expression
function TJSTable.JSGetLookupCursor(const SearchExpr: String): HDBICur;
var
  CompareExpr, IndexExpr, IndexFile, IndexTag: String;
  ExprLen, ItrIdx, IdxIndex: Integer;
begin
  if SearchExpr <> FLookupIndexExpr then
  begin
    JSDestroyLookupCursor;
    CompareExpr := SearchExpr;
    ExprLen := Length(CompareExpr);
    // Remove trailing ')' (for case insensitive expression)
    if (ExprLen > 0) and (CompareExpr[ExprLen] = ')') then
    begin
      System.Delete(CompareExpr, ExprLen, 1);
      Dec(ExprLen);
    end;
    IndexDefs.Update;
    IdxIndex := -1;
    for ItrIdx := 0 to IndexDefs.Count - 1 do
    begin
      with IndexDefs.Items[ItrIdx] do
      begin
        IndexExpr := Fields+Expression;
        if StrLIComp(PChar(IndexExpr), PChar(CompareExpr), ExprLen) = 0 then
        begin
          if Name = IndexName then
          begin
            IdxIndex := ItrIdx;
            Break;
          end else
          if IdxIndex < 0 then
            IdxIndex := ItrIdx;
        end;
      end;
    end;
    if IdxIndex >= 0 then
    begin
      // When the table uses dBase .MDX indexes, Name indicates the index
      // tag name and Source is the name of the index file.
      // Source is applicable to dBase table indexes only.
      with IndexDefs.Items[IdxIndex] do
      begin
        if Source = '' then
        begin
          IndexFile := Name;
          IndexTag := '';
        end else
        begin
          IndexFile := Source;
          IndexTag := Name;
        end;
      end;
      Check(DbiCloneCursor(Handle, True, False, FLookupHandle));
      // DbiSwitchToIndex does nothing if the current index is specified
      Check(DbiSwitchToIndex(FLookupHandle, PChar(IndexFile),
        PChar(IndexTag), 0, False));
    end;
    FLookupIndexExpr := SearchExpr;
  end;
  Result := FLookupHandle;
end;

procedure TJSTable.JSDestroyLookupCursor;
begin
  if FLookupHandle <> nil then
  begin
    DbiCloseCursor(FLookupHandle);
    FLookupHandle := nil;
    FLookupIndexExpr := '';
  end;
end;

procedure TJSTable.InitFieldDefs;
begin
  if FFieldDescs <> nil then
  begin
    FreeMem(FFieldDescs, FFieldDescCount * SizeOf(FLDDesc));
    FFieldDescs := nil;
  end;
  inherited;
end;

procedure TJSTable.DestroyLookupCursor;
begin
  JSDestroyLookupCursor;
  inherited;
end;

// Avoid 'Field not indexed' error when using expression indexes
function TJSTable.GetIsIndexField(Field: TField): Boolean;
begin
  Result := FInLookup or inherited GetIsIndexField(Field);
end;

// Translates a fieldlist into a dBase expression index.
// The most obvious translation for each field type is used.
procedure TJSTable.GetDBaseExprInfo(const Fields: String;
  Options: TLocateOptions; var ExprInfo: TDBaseExprInfo);
var
  Pos: Integer;
  FieldInfo: TDBaseExprInfo;
begin
  ExprInfo.Expression := '';
  ExprInfo.KeyLength := 0;
  Pos := 1;
  while Pos <= Length(Fields) do
  begin
    GetDBaseFieldExprInfo(ExtractFieldName(Fields, Pos), FieldInfo);
    ExprInfo.Expression := ExprInfo.Expression + FieldInfo.Expression + '+';
    ExprInfo.KeyLength := ExprInfo.KeyLength + FieldInfo.KeyLength;
  end;
  if ExprInfo.Expression <> '' then
  begin
    if loPartialKey in Options then
      ExprInfo.KeyLength := ExprInfo.KeyLength - FieldInfo.KeyLength;
    System.Delete(ExprInfo.Expression, Length(ExprInfo.Expression), 1);
    if loCaseInsensitive in Options then
      ExprInfo.Expression := 'UPPER('+ExprInfo.Expression+')';
  end;
end;

procedure TJSTable.GetDBaseFieldExprInfo(const FieldName: String;
  var ExprInfo: TDBaseExprInfo);
const
  InvalidFieldMsg = 'Invalid expression index field type. Field: ''%s''';
var
  FieldDef: TFieldDef;
  FieldDesc: PFldDesc;
begin
  FieldDef := FieldDefs.Find(FieldName);
  case FieldDef.DataType of
    ftString:
    begin
      ExprInfo.Expression := FieldName;
      ExprInfo.KeyLength := FieldDef.Size;
    end;
    ftDate, ftDateTime:
    begin
      ExprInfo.Expression := 'DTOS('+FieldName+')';
      ExprInfo.KeyLength := 8;
    end;
    ftBoolean:
    begin
      ExprInfo.Expression := 'IIF('+FieldName+',"1","0")';
      ExprInfo.KeyLength := 1;
    end;
    ftSmallint, ftWord, ftInteger, ftFloat, ftCurrency, ftBCD:
    begin
      FieldDesc := GetFieldDesc(FieldName);
      if FieldDesc = nil then
        DatabaseErrorFmt(InvalidFieldMsg, [FieldName]);
      ExprInfo.Expression := 'STR('+FieldName+','+
        IntToStr(FieldDesc.iUnits1)+','+
        IntToStr(FieldDesc.iUnits2)+')';
      ExprInfo.KeyLength := FieldDesc.iUnits1;
    end;
  else
    DatabaseErrorFmt(InvalidFieldMsg, [FieldName]);
  end;
end;

// Gets low level BDE information of a field
function TJSTable.GetFieldDesc(FieldName: String): PFldDesc;
var
  CursorProps: CurProps;
  ItrDsc: Integer;
begin
  Result := nil;
  if FFieldDescs = nil then
  begin
    DbiGetCursorProps(Handle, CursorProps);
    FFieldDescCount := CursorProps.iFields;
    FFieldDescs := AllocMem(FFieldDescCount * SizeOf(FLDDesc));
    Check(DbiGetFieldDescs(Handle, PFldDesc(FFieldDescs)));
  end;
  for ItrDsc := 0 to FFieldDescCount - 1 do
    if StrIComp(FFieldDescs^[ItrDsc].szName, PChar(FieldName)) = 0 then
    begin
      Result := @FFieldDescs^[ItrDsc];
      Break;
    end;
end;


procedure Register;
begin
  RegisterComponents('Samples', [TJSTable]);
end;

end.
